/*
 *  linux/drivers/block/wb553.c 
 *
 *  Copyright (C) 1996  Corel Computer Corp
 */

/*
 * The WinBond 533 IDE controller with PCI bus
 * accessed through the Digital 21285 (Footbridge)
 *
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <asm/io.h>
#include "ide.h"
#include "ide_modes.h"

#define DEFAULT_BMIBA	0xE800	/* e800 */

/*
 * PCI config register
 * 40: 1=enable primary ide port
 *     2=enable fast functions on primary ide port
 *     0x10=enable secondary port
 *     0x20=enable fast functions on secondary ide port
 * 44: 0-0x1F=cmd off time	primary port	master
 *     0x20=read-ahead enable	primary port	master
 *     0x40=iochrdy enable	primary port 	master
 *     0x80=posted write enable	primary port	master
 *     0x100-0x1f00=cmd on time primary port	master
 * 48: as above			primary port	slave
 * 4C: as above			secondary port	master
 * 50: as above			secondary port	slave
 *
 * For 33MHz PCI bus, the on and off times should be set to
 *	mode	cycle	on 	off
 *	PIO 0	600n	5	13
 *	PIO 1	383n	4	7
 *	PIO 2	240n	3	4
 *	PIO 3	180n	2	2
 * 	PIO 4	120n	2	0
 *	PIO 5	90n	1	0
 *
 * The PCI config register is at 0xE0500100 = IO_BASE_PCI_CONFIG + 0x00400000 + 0x100	
 *
 */

extern void init_triton_dma (ide_hwif_t *hwif, unsigned short base);

static char pioSpeeds[] = {5,13,	4,7,	3,4,	2,2,	2,0,	1,0};
static char mdmaSpeeds[] = {7,7,	2,1,	2,0,	1,0};
static unsigned short mdma_times[] = {480,	150,	120,	90};

static void tune_wb553 (ide_drive_t *drive, byte pio)
{
	unsigned long flags;
	volatile long* pConfigReg = (long*) (IO_BASE_PCI_CONFIG + 0x00400000 + 0x100);
	struct hd_driveid *id = drive->id;
	unsigned short	dma_mword = id->dma_mword;
	unsigned short cmd_on_time, cmd_off_time;
	long config_speed;
	int dma_mode = 0;

//printk("tune_wb553 about to call get_best, requesting %x\n",pio);
//	pio = ide_get_best_pio_mode(drive, pio, 3, NULL);
//	if (pio>5) pio = 5;
//printk("tune_wb553 to mode %x on drive %d\n",pio,drive->select.b.unit);
//printk("old value of 40: %x 44: %x\n",pConfigReg[0x10],pConfigReg[0x11 + drive->select.b.unit]);

	printk(KERN_DEBUG "%s: tune_wb553\n",drive->name);

	save_flags(flags);
	cli();
	// general ide enable registers
	pConfigReg[0x10] |= 3;

	// set data transfer timing parameters

	if (drive->using_dma)
	{
		dma_mword >>= 8;
		for (; (dma_mode <= 3) && !(dma_mword & 0x01); dma_mode++)
		{
			dma_mword >>= 1;
		}
		
		if (dma_mode > 3)
		{
			dma_mode = 3;
		}

		//if (mdma_times[dma_mode] >= id->eide_dma_time)
		if ( mdma_times[dma_mode] >= id->eide_dma_min && 
				 id->field_valid & 0x02
				)
		{
			config_speed = (mdmaSpeeds[2*dma_mode] << 8) + mdmaSpeeds[2*dma_mode + 1];
		}
		else if (id->field_valid & 0x02) // if words 64...70 are valid
		{
			printk(KERN_DEBUG "%s: DMA times recalculation\n", drive->name);
			// Time calculations from manual for W83C553F
			cmd_on_time = ((id->eide_dma_min/2 + 10) + 29)/30 - 1;
			cmd_off_time = ((id->eide_dma_min - (cmd_on_time + 1)*30 + 29))/30 - 1;
			
			config_speed = (long)(cmd_on_time * 0x100 + cmd_off_time);
		}
		else
		{
			printk(KERN_DEBUG "%s: DMA default mword0\n", drive->name);
			dma_mode = 0;
			config_speed = (mdmaSpeeds[2*dma_mode] << 8) + mdmaSpeeds[2*dma_mode + 1];
		}
	}
	else
	{
		pio = ide_get_best_pio_mode(drive, pio, 5, NULL);

		config_speed = (pioSpeeds[2*pio] << 8) + pioSpeeds[2*pio + 1];
		if (pio > 2)
		{
//----------------------------------------------------------------
			//*
			//* According to ATA-2 spec drives reporting support for PIO 
			//* transfer mode 3 or 4 shall activate PIO mode 0,1 or 2 after
			//* power up or hardware reset. 
			//*

			OUT_BYTE(drive->select.all, IDE_SELECT_REG);
			OUT_BYTE(drive->ctl | 0x02,IDE_CONTROL_REG);	// disable INTRQ
			// set DMA transfer mode based on value	in Sector Count reg.
			OUT_BYTE(0x03,IDE_FEATURE_REG);

			OUT_BYTE(0x08|pio,IDE_NSECTOR_REG);
			OUT_BYTE(WIN_SETFEATURES, IDE_COMMAND_REG);

			if (ide_wait_stat(drive, DRIVE_READY, BUSY_STAT|DRQ_STAT|ERR_STAT, WAIT_CMD)) {
				printk("%s: drive not ready for command\n", drive->name);
//				return 1;
			}

			OUT_BYTE(drive->ctl,IDE_CONTROL_REG);	/* enable device irq */
//----------------------------------------------------------------
			config_speed |= 0x40; // enable IORDY for advanced PIO modes.
		}
	}

	pConfigReg[0x11 + drive->select.b.unit] = config_speed;
	//pConfigReg[0x11 + drive->select.b.unit] = 0x0304;
	restore_flags(flags);

	printk(KERN_DEBUG "%s: %s mode %d\n",drive->name, 
					(drive->using_dma)?"DMA mword":"PIO", (drive->using_dma)?dma_mode:pio);
//printk("new values of reg 40: %x, reg 44: %x\n",pConfigReg[0x10],pConfigReg[0x11 + drive->select.b.unit]);
}

void ide_init_wb553 (void)
{
	unsigned short pcicmd;
	unsigned int ctrl_stat;
	int h;
	char *pConfigReg = (char *)(IO_BASE_PCI_CONFIG + 0x00400000 + 0x100);

	// Check for Vendor & Device IDs 

	// Read device control register.
	pcicmd = *(unsigned short *)(pConfigReg + 0x04);

	if ((pcicmd & 1) == 0)
	{
		printk("ide: ports are not enabled\n");
	}

	/*
	 * See if ide port(s) are enabled
	 */

	ctrl_stat = *(int *)(pConfigReg + 0x40);

	if (!(ctrl_stat & 0x09))
	{
		printk("ide: neither port is enabled\n");
	}

#ifdef CONFIG_BLK_DEV_TRITON
	// Enable Bus Master DMA (set BMEN bit)
	*(pConfigReg + 0x04) |= 0x04;
	// Set Bus Master interface base address
	*(int *)(pConfigReg + 0x20) = DEFAULT_BMIBA;
#endif /* CONFIG_BLK_DEV_TRITON */

	for (h = 0; h < MAX_HWIFS; ++h)
	{
		ide_hwif_t *hwif = &ide_hwifs[h];

		if (
				((hwif->io_base == 0x1f0) && (ctrl_stat & 0x01)) ||
				((hwif->io_base == 0x170) && (ctrl_stat & 0x08))
				)
		{
			hwif->chipset = ide_wb553;
			hwif->tuneproc = &tune_wb553;
#ifdef CONFIG_BLK_DEV_TRITON
			init_triton_dma(hwif, DEFAULT_BMIBA);
#endif /* CONFIG_BLK_DEV_TRITON */
		}
	}
}
