/*
    NetWinder Floating Point Emulator
    (c) Corel Computer Corporation, 1998

    Direct questions, comments to Scott Bambrough <scottb@corelcomputer.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/sched.h>
#include "softfloat.h"
#include "fpopcode.h"
#include "fpmodule.h"
#include "fpa11.h"
#include "fpa11.inl"

/* external data */
extern FPA11 fpa11;

/* kernel symbols required for signal handling */
typedef struct task_struct*	PTASK;
#ifdef MODULE
extern PTASK C_SYMBOL_NAME(current_set);
void C_SYMBOL_NAME(fp_send_sig)(unsigned long sig, PTASK p, int priv);
#endif

/* kernel function prototypes required */
void C_SYMBOL_NAME(fp_setup)(void);

/* external declarations for saved kernel symbols */
extern unsigned int C_SYMBOL_NAME(fp_enter);
extern unsigned int C_SYMBOL_NAME(fp_save);
extern unsigned int C_SYMBOL_NAME(fp_restore);

/* forward declarations */
void C_SYMBOL_NAME(fpenter)(void);
void C_SYMBOL_NAME(fpsave)(unsigned int *);
void C_SYMBOL_NAME(fprestore)(unsigned int *);

/* Original value of fp_enter from kernel before patched by fpe_init. */ 
static unsigned int orig_fp_enter;

/* Original value of fp_save from kernel before patched by fpe_init. */ 
static unsigned int orig_fp_save;

/* Original value of fp_restore from kernel before patched by fpe_init. */ 
static unsigned int orig_fp_restore;

/* Address of user registers on the kernel stack. */
unsigned int *userRegisters;

void C_SYMBOL_NAME(fpe_version)(void)
{
  static const char szTitle[] = "<4>NetWinder Floating Point Emulator ";
  static const char szVersion[] = "V0.94  ";
  static const char szCopyright[] = "(c) 1998 Corel Computer Corp.\n";
  C_SYMBOL_NAME(fp_printk)(szTitle);
  C_SYMBOL_NAME(fp_printk)(szVersion);
  C_SYMBOL_NAME(fp_printk)(szCopyright);
}

int fpe_init(void)
{
  /* Save original values of kernel symbols fp_enter, fp_save & fp_restore. */  
  orig_fp_enter   = C_SYMBOL_NAME(fp_enter);
  orig_fp_save    = C_SYMBOL_NAME(fp_save);
  orig_fp_restore = C_SYMBOL_NAME(fp_restore);
  
  /* Patch kernel symbols fp_enter, fp_save & fp_restore to point
     to routines in the emulator module. */
  C_SYMBOL_NAME(fp_enter)   = (unsigned int)C_SYMBOL_NAME(fpenter);
  C_SYMBOL_NAME(fp_save)    = (unsigned int)C_SYMBOL_NAME(fpsave);
  C_SYMBOL_NAME(fp_restore) = (unsigned int)C_SYMBOL_NAME(fprestore);

  /* Call floating point emulator setup routine in the kernel. */
  C_SYMBOL_NAME(fp_setup)();

  /* Initialize emulator */
  resetFPA11();
  SetRoundingMode(ROUND_TO_NEAREST);
  SetRoundingPrecision(ROUND_EXTENDED);

  /* Display title, version and copyright information. */
  C_SYMBOL_NAME(fpe_version)();
   
  return 0;
}

#ifdef MODULE
int init_module(void)
{
   return(fpe_init());
}

void cleanup_module(void)
{
  /* Restore kernel symbols fp_enter, fp_save & fp_restore to point
     to values saved during init_module. */
  C_SYMBOL_NAME(fp_enter)   = orig_fp_enter;
  C_SYMBOL_NAME(fp_save)    = orig_fp_save;
  C_SYMBOL_NAME(fp_restore) = orig_fp_restore;
}
#endif

void C_SYMBOL_NAME(fpsave)(unsigned int *pBuf)
{
  C_SYMBOL_NAME(memcpy)(pBuf,&fpa11,sizeof(FPA11));
}

void C_SYMBOL_NAME(fprestore)(unsigned int *pBuf)
{
  C_SYMBOL_NAME(memcpy)(&fpa11,pBuf,sizeof(FPA11));
}

/* 
This is the kernel's entry point into the floating point emulator.  It is 
called from the kernel with code similar to this:

.L10	.word	_fp_enter
	mov	r4, .L10
	ldr	r4, [r4]
	blne	_math_state_restore	@restore the FP state if necessary
	adr	r9, _fpreturn		@r9 = address to return to is successful
	adr	lr, _fpundefinstr	@lr = address to return to if undefined instruction
	ldr	pc, [r4]		@jump to the FP emulator's entry point
	
The kernel expects the emulator to return via one of two possible points
of return it passes to the emulator.  The emulator, if successful in its 
emulation, jumps to _fpreturn (passed in r9) and the kernel takes care of 
returning control from the trap to the user code.  If the emulator is unable
to emulate the instruction, it returns via _fpundefinstr (passed via lr) and
the kernel halts the user program with a core dump.

This routine does three things:

1) It saves SP into a variable called userRegisters.  The kernel has created
a struct pt_regs on the stack and saved the user registers into it.  See 
/usr/include/asm/proc/ptrace.h for details.  The emulator code uses
userRegisters as the base of an array of words from which the contents of the
registers can be extracted.

2) It calls EmulateAll to emulate a floating point instruction.  EmulateAll
returns 1 if the emulation was successful, or 0 if not.

3) If an instruction has been emulated successfully, it looks ahead at the
next instruction.  If it is a floating point instruction, it executes the
instruction, without returning to user space.  In this way it repeatedly 
looks ahead and executes floating point instructions until it encounters
a non floating point instruction, at which time it returns via _fpreturn.

This is done to reduce the effect of the trap overhead on each floating
point instructions.  GCC attempts to group floating point instructions to
allow the emulator to spread the cost of the trap over several floating
point instructions.
*/

#define _ARM_pc 60
#define _ARM_cpsr 64

void C_SYMBOL_NAME(fpenter)(void)
{
  /*
     r4 = pointer to user regs
     r5 = program counter
     r6 = current instruction
  */
    __asm__("
            str sp, %0                  @ save pointer to user regs;
            mov r4, sp                  @ use r4 for local pointer
            mov r10, lr  		@ save the failure-return addresses;
   
            ldr r5, [r4, #60]     	@ get contents of PC;
	    ldr r0, [r5, #-4]		@ get actual instruction into r0;
   emulate: bl EmulateAll		@ emulate the instruction;
   	    cmp r0, #0			@ was emulation successful;
            moveq pc, r10      		@ no, return failure

      next: ldr r6, [r5]		@ get the next instruction;

            mov r0, r6                  @ save a copy
            ldr r1, [r4, #64]           @ fetch the condition codes
   	    bl  checkCondition		@ check the condition;
   	    cmp r0, #0			@ r0 = 0 ==> condition failed;

            @ if condition code failed to match, next insn
            addeq r5, r5, #4		@ increment PC;
            streq r5, [r4, #60]         @ update regs
   	    beq next			@ get the next instruction;
   	    
      test: and   r2, r6, #0x0F000000   @ test for FP insns
            teq   r2, #0x0C000000
            teqne r2, #0x0D000000
            teqne r2, #0x0E000000
            movne pc, r9                @ return ok if not a fp insn
   	     			
            add r5, r5, #4		@ increment PC;
            mov r0, r6                  @ prepare for EmulateAll()
            str r5, [r4, #60]           @ userRegs must be up to date for EmulateAll()
   	    b emulate			@ if r0 != 0, goto EmulateAll; 

            "
   	    : /* no outputs */
            : "g" (userRegisters));
}

/*
ScottB:  November 4, 1998

Moved this function out of softfloat-specialize into fpmodule.c.
This effectively isolates all the changes required for integrating with the
Linux kernel into fpmodule.c.  Porting to NetBSD should only require modifying
fpmodule.c to integrate with the NetBSD kernel (I hope!).

float_exception_flags is a global variable in SoftFloat.

This function is called by the SoftFloat routines to raise a floating
point exception.  We check the trap enable byte in the FPSR, and raise
a SIGFPE exception if necessary.  If not the relevant bits in the 
cumulative exceptions flag byte are set and we return.
*/

void float_raise(signed char flags)
{
   //fp_printk("nwfpem: float_raise(0x%08x), FPSR = 0x%08x\n",flags,readFPSR());

   float_exception_flags |= flags;
   if (readFPSR() & (flags << 16))
   {
     /* raise exception */
     //fp_printk("nwfpem: raising exception.\n");
     C_SYMBOL_NAME(fp_send_sig)(SIGFPE,C_SYMBOL_NAME(current_set[0]),1);
   }
   else
   {
     /* set the cumulative exceptions flags */
     //fp_printk("nwfpem: setting cumulative exception flags.\n");
     writeFPSR(flags);
   }
}
